// SPDX-License-Identifier: GPL-2.0+
/* NetworkManager system settings service
 *
 * Based on ifcfg-fedora and ifcfg-suse code
 *
 * (C) Copyright 2008 - 2020 Mikhail Efremov <sem@altlinux.org>
 */

#include "nms-etcnet-alt-common.h"
#include <sys/stat.h>

#include "nms-etcnet-alt-plugin.h"
#include "nms-etcnet-alt-reader.h"
#include "nms-etcnet-alt-shvar.h"
#include "nm-config.h"

#include "nm-core-utils.h"
#include "settings/nm-settings-plugin.h"
#include "settings/nm-settings-storage.h"

#define ETCNET_PATH        SYSCONFDIR "/net"
#define ETCNET_IFACES_PATH ETCNET_PATH "/ifaces/"

typedef struct {
    NMConnection *     connection;
    NMSettingsStorage *storage;
} StorageData;

typedef struct {
    char *      ifaces_path;

    GHashTable * connections_sd;
    GHashTable * unmanaged_specs;

    bool initialized : 1;
} NMSEtcnetPluginPrivate;

struct _NMSEtcnetPlugin {
    NMSettingsPlugin       parent;
    NMSEtcnetPluginPrivate _priv;
};

struct _NMSEtcnetPluginClass {
    NMSettingsPluginClass parent;
};

G_DEFINE_TYPE(NMSEtcnetPlugin, nms_etcnet_plugin, NM_TYPE_SETTINGS_PLUGIN)

#define NMS_ETCNET_PLUGIN_GET_PRIVATE(self) \
    _NM_GET_PRIVATE(self, NMSEtcnetPlugin, NMS_IS_ETCNET_PLUGIN)

static char *
get_iface_path(const char *ifaces_path, const char *iface)
{
    return g_build_filename(ifaces_path, iface, NULL);
}

static const char *
_iface_name(const char *iface_path)
{
    const char *p = NULL;
    g_return_val_if_fail(iface_path != NULL, NULL);

    if (iface_path) {
        p = strrchr(iface_path, '/');
        if (p)
            ++p;
    }

    return p;
}

static GSList *
get_unmanaged_specs(NMSettingsPlugin *config)
{
    GSList *                list = NULL;
    char *                  spec = NULL;
    GHashTableIter          iter;
    NMSEtcnetPlugin *       self = NMS_ETCNET_PLUGIN(config);
    NMSEtcnetPluginPrivate *priv = NMS_ETCNET_PLUGIN_GET_PRIVATE(self);

    g_return_val_if_fail(priv->unmanaged_specs != NULL, NULL);

    g_hash_table_iter_init(&iter, priv->unmanaged_specs);
    while (g_hash_table_iter_next(&iter, NULL, (gpointer *) &spec)) {
        g_assert(spec != NULL);
        list = g_slist_prepend(list, g_strdup(spec));
    }
    return list;
}

static GSList *
read_ifaces_dir(const char *ifaces_dir_name)
{
    GDir *  ifaces_dir;
    const char *  file_name     = NULL;
    GSList * ifaces_list  = NULL;

    ifaces_dir = g_dir_open(ifaces_dir_name, 0, NULL);
    if (!ifaces_dir) {
        return NULL;
    }

    while ((file_name = g_dir_read_name(ifaces_dir))) {
        char *iface_path;

        if (!strcmp("default", file_name) ||
            !strcmp("unknown", file_name) ||
            !strcmp("lo", file_name)) {
            continue;
        }

        iface_path = get_iface_path(ifaces_dir_name, file_name);
        if (g_file_test(iface_path, G_FILE_TEST_IS_DIR)) {
            ifaces_list = g_slist_prepend(ifaces_list, g_strdup(file_name));
        }

        g_free (iface_path);
    }

    g_dir_close(ifaces_dir);

    return ifaces_list;
}

static void
_storage_data_destroy(StorageData *sd)
{
    nm_g_object_unref(sd->connection);
    nm_g_object_unref(sd->storage);
    g_slice_free(StorageData, sd);
}

static NMConnection *
create_etcnet_connection(const char * iface_path,
                         GHashTable * wsconf,
                         gboolean *   ignored)
{
    NMConnection *connection;
    GError *      error           = NULL;

    g_return_val_if_fail(iface_path != NULL, NULL);
    g_return_val_if_fail(ignored != NULL, NULL);

    connection =
        read_connection(iface_path, wsconf, ignored, &error);

    if (!connection) {
        if (!*ignored && error)
            PLUGIN_ERROR("read connection", error);
        goto done;
    }

    if (!nm_connection_normalize(connection, NULL, NULL, &error)) {
        PLUGIN_ERROR("nm_connection_normalize", error);
        g_object_unref(connection);
        connection = NULL;
        goto done;
    }

done:
    if (error)
        g_error_free(error);

    return connection;
}

/* Return TRUE if connection was created or interface is not ignored
 * and FALSE otherwise */
static gboolean
add_one_connection_for_device(NMSEtcnetPlugin *plugin,
                              const char *iface_path,
                              GHashTable *    wsconf)
{
    NMConnection *connection;
    gboolean      ignored;
    const char *       ssid           = NULL;
    char *             connection_key = NULL;
    NMSettingsStorage *storage        = NULL;
    StorageData *      sd;
    const char *       uuid = NULL;
    NMSEtcnetPluginPrivate *priv = NMS_ETCNET_PLUGIN_GET_PRIVATE(plugin);

    g_return_val_if_fail(iface_path != NULL, FALSE);
    g_return_val_if_fail(priv->connections_sd != NULL, FALSE);

    connection = create_etcnet_connection(iface_path,
                                          wsconf,
                                          &ignored);
    if (ignored) {
        if (connection) {
            PLUGIN_WARNING("Ignored, but readed connection");
            g_object_unref(connection);
        }
        return FALSE;
    }

    if (connection == NULL) {
        return ignored ? FALSE : TRUE;
    }

    if (wsconf)
        ssid = (const char *) g_hash_table_lookup(wsconf, "ssid");

    connection_key = ssid ? g_strdup_printf("%s(%s)", iface_path, ssid) : g_strdup(iface_path);

    nm_assert_connection_unchanging(connection);
    uuid = nm_connection_get_uuid(connection);

    storage = nm_settings_storage_new(NM_SETTINGS_PLUGIN(plugin), uuid, NULL);

    sd  = g_slice_new(StorageData);
    *sd = (StorageData){
        .connection = connection,
        .storage    = storage,
    };

    g_hash_table_insert(priv->connections_sd, connection_key, sd);
    PLUGIN_INFO("%s: Readed connection \"%s\"",
                _iface_name(iface_path),
                nm_connection_get_id(connection));

    return TRUE;
}

static gboolean
add_connections_for_device(NMSEtcnetPlugin *plugin, const char *iface_path)
{
    GHashTable *wsconf        = NULL;
    GSList *    wsconfig_list = NULL;
    GSList *    list_item;
    guint       num_readed = 0;

    g_return_val_if_fail(plugin != NULL, FALSE);
    g_return_val_if_fail(iface_path != NULL, FALSE);

    list_item = wsconfig_list = read_supplicant_config(iface_path);
    do {
        if (list_item)
            wsconf = (GHashTable *) list_item->data;
        if (add_one_connection_for_device(plugin, iface_path, wsconf))
            ++num_readed;
        list_item = g_slist_next(list_item);
    } while (list_item);
    clear_supplicant_config(wsconfig_list);

    return num_readed > 0;
}

static void
read_device_config(NMSEtcnetPlugin *   plugin,
                   const char      *   iface_name,
                   guint *             out_unmanaged_changed)
{
    char *iface_path;
    gboolean result;
    NMSEtcnetPluginPrivate *priv = NMS_ETCNET_PLUGIN_GET_PRIVATE(plugin);

    g_return_if_fail(plugin != NULL);
    g_return_if_fail(iface_name != NULL);

    iface_path = get_iface_path(priv->ifaces_path, iface_name);
    if (iface_path == NULL)
        return;

    result = add_connections_for_device(plugin, iface_path);

    g_free(iface_path);

    if (result) {
        if (g_hash_table_remove(priv->unmanaged_specs, iface_name))
                (*out_unmanaged_changed)++;
    } else {
        char *spec = NULL;

        spec = g_strdup_printf(NM_MATCH_SPEC_INTERFACE_NAME_TAG "%s", iface_name);

        if (spec) {
            if (g_hash_table_insert(priv->unmanaged_specs, g_strdup(iface_name), spec))
                (*out_unmanaged_changed)++;

            PLUGIN_INFO("Device %s is unmanaged", iface_name);
        }
    }
}

static void
read_connections(NMSEtcnetPlugin *plugin)
{
    NMSEtcnetPluginPrivate *priv = NMS_ETCNET_PLUGIN_GET_PRIVATE(plugin);
	GSList                 *ifaces_list, *iter;
    guint                   unmanaged_changed = 0;

    /* Reset all connections
	 * May be we must update connections, not reset them
	 * But is this really worth the efforts?
	 */
    if (priv->connections_sd) {
        g_hash_table_destroy(priv->connections_sd);
        priv->connections_sd = NULL;
    }

    priv->connections_sd = g_hash_table_new_full(nm_str_hash,
                                                 g_str_equal,
                                                 g_free,
                                                 (GDestroyNotify) _storage_data_destroy);

	ifaces_list = read_ifaces_dir(priv->ifaces_path);
    if (ifaces_list == NULL)
        return;

    for (iter = ifaces_list; iter; iter = g_slist_next(iter)) {
        char *iface_name = (char *) iter->data;
        read_device_config(plugin, iface_name, &unmanaged_changed);
    }

    if (ifaces_list)
        g_slist_free_full(ifaces_list, (GDestroyNotify) g_free);

    if (unmanaged_changed)
        _nm_settings_plugin_emit_signal_unmanaged_specs_changed(NM_SETTINGS_PLUGIN(plugin));
}

static void
initialize(NMSEtcnetPlugin *self)
{
    NMSEtcnetPluginPrivate *priv = NMS_ETCNET_PLUGIN_GET_PRIVATE(self);

    nm_assert(!priv->initialized);

    priv->initialized = TRUE;
}

static void
reload_connections(NMSettingsPlugin *                     plugin,
                   NMSettingsPluginConnectionLoadCallback callback,
                   gpointer                               user_data)
{
    NMSEtcnetPlugin *       self = NMS_ETCNET_PLUGIN(plugin);
    NMSEtcnetPluginPrivate *priv = NMS_ETCNET_PLUGIN_GET_PRIVATE(self);
    GHashTableIter          iter_sd;
    StorageData *           sd  = NULL;

    if (!priv->initialized) {
        initialize(self);
    }

    if (priv->connections_sd) {
        g_hash_table_iter_init(&iter_sd, priv->connections_sd);
        while (g_hash_table_iter_next(&iter_sd, NULL, (gpointer *) &sd)) {
            if (!sd)
                continue;
            PLUGIN_DEBUG("unload: %s", nm_settings_storage_get_uuid(sd->storage));
            callback(plugin, sd->storage, NULL, user_data);
        }
    }

    read_connections(self);

    g_hash_table_iter_init(&iter_sd, priv->connections_sd);
    while (g_hash_table_iter_next(&iter_sd, NULL, (gpointer *) &sd)) {
        if (!sd)
            continue;

        PLUGIN_DEBUG("load: %s (%s)",
                     nm_settings_storage_get_uuid(sd->storage),
                     nm_connection_get_id(sd->connection));
        callback(plugin, sd->storage, sd->connection, user_data);
    }
}

static void
nms_etcnet_plugin_init(NMSEtcnetPlugin *plugin)
{
    NMSEtcnetPluginPrivate *priv      = NMS_ETCNET_PLUGIN_GET_PRIVATE(plugin);

    priv->ifaces_path = g_strdup(ETCNET_IFACES_PATH);
    if (!priv->ifaces_path)
        PLUGIN_WARNING("could not determine network ifaces path.");
    priv->connections_sd = NULL;
    priv->unmanaged_specs = g_hash_table_new_full(nm_str_hash, g_str_equal, g_free, g_free);;
}

static void
dispose(GObject *object)
{
    NMSEtcnetPlugin *       plugin = NMS_ETCNET_PLUGIN(object);
    NMSEtcnetPluginPrivate *priv   = NMS_ETCNET_PLUGIN_GET_PRIVATE(plugin);

    g_free(priv->ifaces_path);

    if (priv->connections_sd) {
        g_hash_table_destroy(priv->connections_sd);
    }
    if (priv->unmanaged_specs) {
        g_hash_table_destroy(priv->unmanaged_specs);
    }

    G_OBJECT_CLASS(nms_etcnet_plugin_parent_class)->dispose(object);
}

static void
nms_etcnet_plugin_class_init(NMSEtcnetPluginClass *klass)
{
    GObjectClass *         object_class = G_OBJECT_CLASS(klass);
    NMSettingsPluginClass *plugin_class = NM_SETTINGS_PLUGIN_CLASS(klass);

    object_class->dispose = dispose;

    plugin_class->plugin_name         = "etcnet-alt";
    plugin_class->reload_connections  = reload_connections;
    plugin_class->get_unmanaged_specs = get_unmanaged_specs;
}

G_MODULE_EXPORT NMSettingsPlugin *
                nm_settings_plugin_factory(void)
{
    return g_object_new(NMS_TYPE_ETCNET_PLUGIN, NULL);
}
