// SPDX-License-Identifier: GPL-2.0+
/* NetworkManager system settings service
 *
 * Based on ifcfg-fedora and ifcfg-suse code
 *
 * (C) Copyright 2008 - 2020 Mikhail Efremov <sem@altlinux.org>
 */

#include "nms-etcnet-alt-common.h"
#include <ctype.h>
#include <net/if.h>
#include "NetworkManagerUtils.h"
#include "libnm-platform/nm-platform.h"
#include "libnm-glib-aux/nm-uuid.h"
/* For INT32_MAX */
#include <stdint.h>

#include "nms-etcnet-alt-shvar.h"
#include "nms-etcnet-alt-reader.h"

static shvarFile *
svNewFile_iface_file(const char *iface_path, const char *file)
{
    char      *file_path;
    shvarFile *shvar_f = NULL;

    file_path        = g_build_filename(iface_path, file, NULL);
    shvar_f          = svNewFile(file_path);

    g_free(file_path);

    return shvar_f;
}

typedef struct options_file_t {
    shvarFile *options;
    shvarFile *default_options_devtype;
    shvarFile *default_options;
    char      *type;
} options_file_t;

static char *
options_get_value(options_file_t *of, const char *key)
{
    char *value = NULL;

    g_return_val_if_fail(of != NULL, NULL);
    g_return_val_if_fail(key != NULL, NULL);

    if (of->options)
        value = svGetValue(of->options, key);
    if (!value && of->default_options_devtype)
        value = svGetValue(of->default_options_devtype, key);
    if (!value && of->default_options)
        value = svGetValue(of->default_options, key);

    return value;
}

static gboolean
options_true_value(options_file_t *of, const char *key, gboolean def)
{
    int ret = -1;

    g_return_val_if_fail(of != NULL, def);
    g_return_val_if_fail(key != NULL, def);

    if (of->options)
        ret = svTrueValue(of->options, key, -1);
    if (ret < 0 && of->default_options_devtype)
        ret = svTrueValue(of->default_options_devtype, key, -1);
    if (ret < 0)
        ret = of->default_options ? svTrueValue(of->default_options, key, def) : def;

    return ret > 0;
}

static char *
get_etcnet_type(options_file_t *options, const char *iface_name)
{
    char *       etcnet_type_val = NULL;

    g_return_val_if_fail(options != NULL, NULL);
    g_return_val_if_fail(iface_name != NULL, NULL);

    etcnet_type_val = options_get_value(options, "TYPE");

    if (etcnet_type_val == NULL) {
        /* etcnet guess type from the interface name.
		 * So do the same, dropping digits at the end of the name */
        size_t tlen = strcspn(iface_name, "0123456789");
        if (tlen > 0 && tlen < strlen(iface_name))
            etcnet_type_val = g_strndup(iface_name, tlen);
    } else
        etcnet_type_val = g_strchomp(etcnet_type_val);

    return etcnet_type_val;
}

static void
close_options_file(options_file_t *of)
{
    if (!of)
        return;

    if (of->options)
        svCloseFile(of->options);
    if (of->default_options_devtype)
        svCloseFile(of->default_options_devtype);
    if (of->default_options)
        svCloseFile(of->default_options);
    if (of->type)
        g_free(of->type);
    g_free(of);
}

static options_file_t *
open_options_file(const char *iface_path)
{
    options_file_t *of;
    char *          ifaces_dir           = NULL;
    char *          iface_name           = NULL;
    char *          def_opt_path         = NULL;
    char *          def_opt_devtype_path = NULL;

    of = g_malloc0(sizeof(options_file_t));
    if (!of) {
        PLUGIN_WARNING("%s: Not enough memory for options_file_t", __func__);
        goto error;
    }
    of->options = svNewFile_iface_file(iface_path, "options");

    ifaces_dir = g_path_get_dirname(iface_path);
    if (!ifaces_dir)
        goto error;

    iface_name = get_iface_name(iface_path);
    if (!iface_name)
        goto error;

    of->type = get_etcnet_type(of, iface_name);
    if (of->type) {
        char opt_devtype_name[IFNAMSIZ + 8];
        gint len;

        len = g_snprintf(opt_devtype_name, sizeof(opt_devtype_name), "options-%s", of->type);
        if (len < sizeof(opt_devtype_name)) {
            def_opt_devtype_path = g_build_filename(ifaces_dir, "default", opt_devtype_name, NULL);
            if (!def_opt_devtype_path)
                goto error;

            of->default_options_devtype = svNewFile(def_opt_devtype_path);
        } else
            PLUGIN_WARNING("Device type name is too long: %s", of->type);
    }

    def_opt_path = g_build_filename(ifaces_dir, "default", "options", NULL);
    if (!def_opt_path)
        goto error;

    of->default_options = svNewFile(def_opt_path);

    g_free(ifaces_dir);
    g_free(iface_name);
    g_free(def_opt_devtype_path);
    g_free(def_opt_path);

    return of;

error:
    g_free(ifaces_dir);
    g_free(iface_name);
    g_free(def_opt_devtype_path);
    g_free(def_opt_path);
    close_options_file(of);

    return NULL;
}

static char *
read_file_contents(const char *iface_path, const char *file, GError **error)
{
    char *  file_fullname;
    char *  contents = NULL;

    g_return_val_if_fail(iface_path != NULL, NULL);
    g_return_val_if_fail(file != NULL, NULL);

    file_fullname    = g_build_filename(iface_path, file, NULL);

    g_file_get_contents(file_fullname, &contents, NULL, error);

    g_free(file_fullname);

    return contents;
}

static char **
read_file_lines(const char *iface_path, const char *file, GError **error)
{
    char **lines    = NULL;
    char * contents = NULL;

    g_return_val_if_fail(iface_path != NULL, NULL);
    g_return_val_if_fail(file != NULL, NULL);

    contents = read_file_contents(iface_path, file, error);
    if (contents) {
        lines = g_strsplit(contents, "\n", 0);
    }
    g_free(contents);

    return lines;
}

static gboolean
get_int(const char *str, long int *value)
{
    char *e;

    g_return_val_if_fail(str != NULL, FALSE);
    g_return_val_if_fail(value != NULL, FALSE);

    *value = strtol(str, &e, 0);
    if (*e != '\0')
        return FALSE;

    return TRUE;
}

static void
strip_line(char *line)
{
    char *   p, *tmp;
    size_t   len;
    int      i, j;
    gboolean delim  = FALSE;
    gboolean quoted = FALSE;

    g_return_if_fail(line != NULL);

    p = strchr(line, '#');
    if (p)
        *p = '\0';
    g_strstrip(line);

    /* Drop duplicate delimiters, but
	 * don't touch delimeters in quotes */
    len = strlen(line);

    tmp = g_malloc(len + 1);
    if (!tmp)
        return;

    for (i = 0, j = 0; i < len; i++) {
        if ((line[i] == ' ' || line[i] == '\t') && !quoted) {
            if (delim)
                continue;
            else
                delim = TRUE;
        } else
            delim = FALSE;

        if (line[i] == '\'' || line[i] == '\"')
            quoted = !quoted;

        tmp[j++] = line[i];
    }
    tmp[j] = '\0';

    if (j != len)
        g_strlcpy(line, tmp, len + 1);

    g_free(tmp);
}

typedef void (*strarray_callback_func)(const char *str, gpointer user_data);

/*
 * Call func for each line in str_array. Lines CAN BE modifyed.
 */
static void
str_array_foreach(char **str_array, strarray_callback_func func, gpointer user_data)
{
    char **pline;

    g_return_if_fail(str_array != NULL);
    g_return_if_fail(func != NULL);

    for (pline = str_array; *pline; ++pline) {
        strip_line(*pline);
        if (**pline != '\0')
            func(*pline, user_data);
    }
}

static gboolean
for_each_line(const char *           iface_path,
              const char *           file,
              strarray_callback_func func,
              gpointer               user_data,
              GError **              error)
{
    char **lines = NULL;

    g_return_val_if_fail(func != NULL, FALSE);

    lines = read_file_lines(iface_path, file, error);
    if (!lines)
        return FALSE;

    str_array_foreach(lines, func, user_data);

    return TRUE;
}

static GHashTable *
read_supplicant_network(char ***plines)
{
    char **     lines   = *plines;
    GHashTable *network = NULL;
    char **     line;
    char **     values;

    g_return_val_if_fail(lines != NULL, NULL);

    network = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
    if (!network) {
        PLUGIN_WARNING("%s: Not enough memory for GHashTable", __func__);
        return NULL;
    }

    for (line = lines + 1; *line; ++line) {
        char *p;

        strip_line(*line);
        if (**line == '\0')
            continue;
        if (**line == '}')
            break;
        values = g_strsplit(*line, "=", 2);
        if (!values || !values[0] || !values[1])
            break;

        if (values[1][0] == '\"') {
            int len;
            p                  = values[1] + 1;
            len                = strlen(values[1]);
            values[1][len - 1] = '\0';
        } else
            p = values[1];

        g_hash_table_insert(network, g_strdup(values[0]), g_strdup(p));
        g_strfreev(values);
    }
    *plines = line;

    if (**line != '}') {
        PLUGIN_WARNING("Not terminated network block!");
        g_hash_table_destroy(network);
        network = NULL;
    }

    return network;
}

GSList *
read_supplicant_config(const char *iface_path)
{
    GSList *    wsconf_list = NULL;
    GHashTable *wsconf      = NULL;
    char **     lines       = NULL;
    char **     pline;

    lines = read_file_lines(iface_path, "wpa_supplicant.conf", NULL);
    if (!lines) {
        goto out;
    }

    for (pline = lines; *pline; ++pline) {
        strip_line(*pline);
        if (**pline == '\0')
            continue;
        if (!strcmp("network={", *pline)) {
            wsconf = read_supplicant_network(&pline);
            if (!wsconf) {
                PLUGIN_WARNING("Parsing wpa_supplicant.conf error");
                break;
            } else {
                wsconf_list = g_slist_prepend(wsconf_list, wsconf);
            }
        }
    }

out:

    g_strfreev(lines);
    return wsconf_list;
}

void
clear_supplicant_config(GSList *wsconfig_list)
{
    if (!wsconfig_list)
        return;

    g_slist_free_full(wsconfig_list, (GDestroyNotify) g_hash_table_destroy);
}

static NMSetting *
make_connection_setting(const char *    iface_path,
                        options_file_t *options,
                        const char *    type,
                        const char *    suggested)
{
    NMSettingConnection *s_con;
    char *               iface_name = NULL;
    char *               new_id = NULL, *uuid = NULL;
    char *               str_for_uuid = NULL;
    gboolean             autoconnect = FALSE;

    g_return_val_if_fail(iface_path != NULL, NULL);
    g_return_val_if_fail(options != NULL, NULL);
    g_return_val_if_fail(type != NULL, NULL);

    iface_name = get_iface_name(iface_path);
    if (!iface_name)
        return NULL;

    s_con = NM_SETTING_CONNECTION(nm_setting_connection_new());

    if (suggested) {
        /* For cosmetic reasons, if the suggested name is the same as
		 * the interface name, don't use it.
		 */
        if (strcmp(iface_name, suggested)) {
            new_id = g_strdup_printf("System %s (%s)",
                                     suggested,
                                     iface_name);
        }
    }

    if (!new_id) {
        new_id = g_strdup_printf("System %s", iface_name);
    }

    g_object_set(s_con,
                 NM_SETTING_CONNECTION_ID,
                 new_id,
                 NM_SETTING_CONNECTION_TYPE,
                 type,
                 NM_SETTING_CONNECTION_READ_ONLY,
                 TRUE,
                 NULL);

    /* In case of suggested == NULL it is just like g_strdup().
	 * Otherwise suggested will be appended.
	 */
    str_for_uuid =
        g_strconcat(options->options ? options->options->fileName : iface_path, suggested, NULL);

    uuid = nm_uuid_generate_from_string_str(str_for_uuid, -1, NM_UUID_TYPE_LEGACY, NULL);
    g_object_set(s_con, NM_SETTING_CONNECTION_UUID, uuid, NULL);

    g_free(str_for_uuid);
    g_free(uuid);
    g_free(new_id);

    /* Be somewhat conservative about autoconnect, */
    autoconnect = options_true_value(options, "ONBOOT", FALSE);
    g_object_set(s_con, NM_SETTING_CONNECTION_AUTOCONNECT, autoconnect, NULL);

    g_free(iface_name);
    return (NMSetting *) s_con;
}

typedef struct iplink_info_t {
    char *mac;
    char *mtu;
} iplink_info_t;

static iplink_info_t *
iplink_info_alloc(void)
{
    return g_malloc0(sizeof(iplink_info_t));
}

static void
iplink_info_free(iplink_info_t *info)
{
    if (!info)
        return;

    g_free(info->mac);
    g_free(info->mtu);
    g_free(info);
}

static iplink_info_t *
get_iplink_info(const char *iface_path, GError **error)
{
    iplink_info_t *info = NULL;
    char **        lines;
    char **        pline;

    info = iplink_info_alloc();

    lines = read_file_lines(iface_path, "iplink", error);
    if (!lines) {
        if (g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
            g_clear_error(error);
        }

        return info;
    }

    for (pline = lines; *pline; ++pline) {
        char **values;
        strip_line(*pline);
        if (**pline == '\0')
            continue;
        values = g_strsplit_set(*pline, " \t", 0);
        if (!values)
            continue;
        if (values[0] && values[1]) {
            g_strstrip(values[0]);
            g_strstrip(values[1]);
            if (!g_ascii_strcasecmp(values[0], "address")) {
                info->mac = g_strdup(values[1]);
            } else if (!g_ascii_strcasecmp(values[0], "mtu")) {
                info->mtu = g_strdup(values[1]);
            }
        }
        g_strfreev(values);
    }
    g_strfreev(lines);

    return info;
}

/***********************************************/
/* Common IP functions and types               */
/***********************************************/
typedef struct ip_conf_cb_info_t {
    gpointer data;
    int      family;
} ip_conf_cb_info_t;

typedef struct add_routes_info_t {
    GSList **p_routes_list;
    char *   default_gateway;
} add_routes_info_t;

/* Get IP address and prefix from string */
static gboolean
split_ipaddr_with_prefix_str(int family, const char *str, char **addr, guint *prefix)
{
    gchar ** str_array = NULL;
    gboolean ret       = FALSE;
    long int max_prefix;
    long int tmp;

    g_return_val_if_fail(str != NULL, FALSE);
    g_return_val_if_fail(addr != NULL, FALSE);
    g_return_val_if_fail(prefix != NULL, FALSE);

    switch (family) {
    case AF_INET:
        max_prefix = 32;
        break;
    case AF_INET6:
        max_prefix = 128;
        break;
    default:
        PLUGIN_WARNING("Invalid family: %d", family);
        return FALSE;
    }

    str_array = g_strsplit(str, "/", 0);

    if (str_array && str_array[0] && str_array[1] && nm_utils_ipaddr_valid(family, str_array[0])
        && (*addr = g_strdup(str_array[0])) && get_int(str_array[1], &tmp) && tmp >= 0
        && tmp <= max_prefix) {
        *prefix = (guint) tmp;
        ret     = TRUE;
    }

    g_strfreev(str_array);

    return ret;
}

static gboolean
set_iproute(NMIPRoute **route, char **default_gateway, int family, const char *line, GError **error)
{
    gchar ** str_array = NULL;
    gboolean ret       = FALSE;
    int      i, gw_index = -1;
    gboolean def_gw      = FALSE;
    guint32  metric      = 0;
    char *   dest_addr   = NULL;
    guint    dest_prefix = 0;

    g_return_val_if_fail(route != NULL, FALSE);
    g_return_val_if_fail(default_gateway != NULL, FALSE);
    g_return_val_if_fail(line != NULL, FALSE);

    str_array = g_strsplit_set(line, " \t", 0);
    if (!str_array || !str_array[0] || !str_array[1]) {
        PLUGIN_SET_ERROR(error, "Invalid IP route string: '%s'", line);
        goto done;
    }

    /* Is it default gateway? */
    if (g_ascii_strcasecmp(str_array[0], "default") == 0) {
        /* Was default gateway configured already? */
        if (*default_gateway != NULL) {
            PLUGIN_SET_ERROR(error, "Can't set multiple default gateways: %s", str_array[0]);
            goto done;
        }

        def_gw = TRUE;
    }

    /* 'via' can not be first */
    for (i = 1; str_array[i] != NULL; ++i) {
        /* search word 'via' in string, next - IP address */
        if (!g_ascii_strcasecmp(str_array[i], "via")) {
            if (str_array[++i] == NULL)
                break;
            gw_index = i;
            continue;
        }
        /* find metric value in string */
        if (!g_ascii_strcasecmp(str_array[i], "metric")) {
            long int tmp;

            if (str_array[++i] == NULL)
                break;
            if (get_int(str_array[i], &tmp) && tmp >= 0 && tmp <= INT32_MAX)
                metric = (guint32) tmp;
            else
                PLUGIN_WARNING("Invalid metric: %s", str_array[i]);
        }
    }

    if (gw_index == -1) {
        PLUGIN_SET_ERROR(error, "Invalid IP route string: '%s'", line);
        goto done;
    }

    if (!nm_utils_ipaddr_valid(family, str_array[gw_index])) {
        PLUGIN_SET_ERROR(error, "Invalid gateway '%s'", str_array[gw_index]);
        goto done;
    }

    if (def_gw) {
        /* Return default gateway */
        *default_gateway = g_strdup(str_array[gw_index]);
        if (*default_gateway == NULL) {
            PLUGIN_SET_ERROR_LITERAL(error, "No memory for default gateway");
            goto done;
        }
    } else {
        /* Return route */
        if (split_ipaddr_with_prefix_str(family, str_array[0], &dest_addr, &dest_prefix)) {
            *route =
                nm_ip_route_new(family, dest_addr, dest_prefix, str_array[gw_index], metric, error);
            if (*route == NULL)
                goto done;

        } else {
            PLUGIN_SET_ERROR(error, "Invalid route: '%s'", line);
            goto done;
        }
    }

    ret = TRUE;
done:
    g_free(dest_addr);
    g_strfreev(str_array);

    return ret;
}

static void
ip_settings_add_address_cb(gpointer data, gpointer user_data)
{
    NMIPAddress *      addr = (NMIPAddress *) data;
    NMSettingIPConfig *s_ip = (NMSettingIPConfig *) user_data;

    g_return_if_fail(addr != NULL);
    g_return_if_fail(s_ip != NULL);

    nm_setting_ip_config_add_address(s_ip, addr);
}

static void
ip_settings_add_route_cb(gpointer data, gpointer user_data)
{
    NMIPRoute *        route = (NMIPRoute *) data;
    NMSettingIPConfig *s_ip  = (NMSettingIPConfig *) user_data;

    g_return_if_fail(route != NULL);
    g_return_if_fail(s_ip != NULL);

    nm_setting_ip_config_add_route(s_ip, route);
}

static void
set_ip_conf_dns_cb(const char *line, gpointer user_data)
{
    ip_conf_cb_info_t *info = (ip_conf_cb_info_t *) user_data;
    NMSettingIPConfig *s_ip;
    int                other_family;
    char **            str_array = NULL;
    guint              array_size;

    g_return_if_fail(info != NULL);
    g_return_if_fail(info->data != NULL);

    s_ip = (NMSettingIPConfig *) info->data;

    /* Fortunately we should handle only 2 families.
	 * If it is not IPv4, then IPv6 */
    switch (info->family) {
    case AF_INET:
        other_family = AF_INET6;
        break;
    case AF_INET6:
        other_family = AF_INET;
        break;
    default:
        /* Never shouldn't happend */
        PLUGIN_WARNING("Invalid IP protocol family: %d", info->family);
        return;
    }

    str_array = g_strsplit_set(line, " \t", 0);
    if (!str_array)
        return;

    array_size = g_strv_length(str_array);
    /* Should be at least 2 tokens */
    if (array_size < 2)
        goto done;

    if (!strcmp(str_array[0], "search") || !strcmp(str_array[0], "domain")) {
        char **item;
        /* Skip first element in array. */
        for (item = str_array + 1; *item; item++)
            nm_setting_ip_config_add_dns_search(s_ip, *item);
    } else if (array_size == 2 && !strcmp(str_array[0], "nameserver")) {
        /* Second element is IP address of the nameserver. */
        if (nm_utils_ipaddr_valid(info->family, str_array[1]))
            nm_setting_ip_config_add_dns(s_ip, str_array[1]);
        else if (nm_utils_ipaddr_valid(other_family, str_array[1])) {
            /* Just skip addresses for another IP protocol family */
            goto done;
        } else
            PLUGIN_WARNING("Invalid DNS server address '%s'", str_array[1]);
    }
done:
    if (str_array)
        g_strfreev(str_array);
}

/*
 * Fill list with IP addresses.
 */
static void
add_ip_conf_addr_cb(const char *line, gpointer user_data)
{
    NMIPAddress *      addr     = NULL;
    char *             addr_str = NULL;
    guint              prefix   = 0;
    GError *           error    = NULL;
    GSList **          list;
    ip_conf_cb_info_t *info = (ip_conf_cb_info_t *) user_data;

    g_return_if_fail(info != NULL);
    g_return_if_fail(info->data != NULL);

    list = (GSList **) info->data;

    if (!split_ipaddr_with_prefix_str(info->family, line, &addr_str, &prefix)) {
        PLUGIN_WARNING("Invalid IP address and prefix string: '%s'", line);
        goto done;
    }

    addr = nm_ip_address_new(info->family, addr_str, prefix, &error);
    if (addr)
        *list = g_slist_prepend(*list, addr);
    else {
        PLUGIN_ERROR("Add IP address", error);
        g_error_free(error);
    }

done:
    g_free(addr_str);
}

/*
 * Fill list with  routes and set default gateway
 */
static void
add_ip_conf_route_cb(const char *line, gpointer user_data)
{
    NMIPRoute *        route = NULL;
    GError *           error = NULL;
    ip_conf_cb_info_t *info  = (ip_conf_cb_info_t *) user_data;
    add_routes_info_t *routes_info;

    g_return_if_fail(info != NULL);
    g_return_if_fail(info->data != NULL);

    routes_info = (add_routes_info_t *) info->data;

    g_return_if_fail(routes_info->p_routes_list != NULL);

    if (set_iproute(&route, &routes_info->default_gateway, info->family, line, &error)) {
        /* Add route to list if it is not default gateway. */
        if (route)
            *routes_info->p_routes_list = g_slist_prepend(*routes_info->p_routes_list, route);
    } else {
        PLUGIN_ERROR("Add route", error);
        g_error_free(error);
    }
}

/***********************************************/
/* IPv4 functions                              */
/***********************************************/
static NMSetting *
make_ip4_setting(options_file_t *options,
                 const char *    iface_path,
                 GError **       error)
{
    NMSettingIPConfig *s_ip4           = NULL;
    char *             value           = NULL;
    char *             method          = NM_SETTING_IP4_CONFIG_METHOD_AUTO;
    GSList *           ip4addr_list    = NULL;
    GSList *           ip4routes_list  = NULL;
    char *             default_gateway = NULL;
    add_routes_info_t  routes_info;
    ip_conf_cb_info_t  info = {NULL, AF_INET};
    gboolean           config_ipv4;

    g_return_val_if_fail(options != NULL, NULL);
    g_return_val_if_fail(iface_path != NULL, NULL);

    config_ipv4 = options_true_value(options, "CONFIG_IPV4", TRUE);
    if (!config_ipv4) {
        method = NM_SETTING_IP4_CONFIG_METHOD_DISABLED;
        goto done;
    }

    value = options_get_value(options, "BOOTPROTO");
    if (value) {
        if (!g_ascii_strncasecmp(value, "static", 6) ||
            /* Yes, it is really static configuration for IPv4
				 * according with that brain-damaged logic in etcnet */
            !g_ascii_strncasecmp(value, "dhcp6", 5))
            method = NM_SETTING_IP4_CONFIG_METHOD_MANUAL;
        else if (!g_ascii_strncasecmp(value, "ipv4ll", 6))
            method = NM_SETTING_IP4_CONFIG_METHOD_LINK_LOCAL;
    }
    g_free(value);

    if (strcmp(method, NM_SETTING_IP4_CONFIG_METHOD_MANUAL) != 0)
        goto done;

    info.data = &ip4addr_list;
    if (for_each_line(iface_path, "ipv4address", add_ip_conf_addr_cb, &info, error)
        && ip4addr_list) {
        /* Reverse the list for matching with order in config file. */
        ip4addr_list = g_slist_reverse(ip4addr_list);
    } else if (g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
        /* Don't care if ipv4address file does not exist */
        g_clear_error(error);
        goto error;
    }

    routes_info.p_routes_list   = &ip4routes_list;
    routes_info.default_gateway = NULL;
    info.data                   = &routes_info;
    if (for_each_line(iface_path, "ipv4route", add_ip_conf_route_cb, &info, error)) {
        /* Reverse the list for matching with order in config file. */
        if (ip4routes_list)
            ip4routes_list = g_slist_reverse(ip4routes_list);
        default_gateway = routes_info.default_gateway;
    } else if (g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
        /* Don't care if ipv4route file does not exist */
        g_clear_error(error);
    } else
        goto error;

done:
    s_ip4 = (NMSettingIPConfig *) nm_setting_ip4_config_new();
    if (!s_ip4) {
        PLUGIN_SET_ERROR_LITERAL(error, "Not enough memory for NMSettingIP4Config");
        goto error;
    }

    g_object_set(s_ip4, NM_SETTING_IP_CONFIG_METHOD, method, NULL);

    if (config_ipv4) {
        if (ip4addr_list) {
            g_slist_foreach(ip4addr_list, ip_settings_add_address_cb, s_ip4);
            g_slist_free_full(ip4addr_list, (GDestroyNotify) nm_ip_address_unref);
            ip4addr_list = NULL;
        }

        if (default_gateway) {
            g_object_set(s_ip4, NM_SETTING_IP_CONFIG_GATEWAY, default_gateway, NULL);
            g_free(default_gateway);
            default_gateway = NULL;
        }

        if (ip4routes_list) {
            g_slist_foreach(ip4routes_list, ip_settings_add_route_cb, s_ip4);
            g_slist_free_full(ip4routes_list, (GDestroyNotify) nm_ip_route_unref);
            ip4routes_list = NULL;
        }

        /* Read resolv.conf */
        info.data = s_ip4;
        if (!for_each_line(iface_path, "resolv.conf", set_ip_conf_dns_cb, &info, error)
            && g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
            /* Don't care if resolv.conf file does not exist */
            g_clear_error(error);
        }
    }

    return NM_SETTING(s_ip4);

error:
    if (ip4addr_list)
        g_slist_free_full(ip4addr_list, (GDestroyNotify) nm_ip_address_unref);
    if (default_gateway)
        g_free(default_gateway);
    if (ip4routes_list)
        g_slist_free_full(ip4routes_list, (GDestroyNotify) nm_ip_route_unref);
    if (s_ip4)
        g_object_unref(s_ip4);
    return NULL;
}

/***********************************************/
/* IPv6 functions                              */
/***********************************************/
static NMSetting *
make_ip6_setting(options_file_t *options,
                 const char *    iface_path,
                 GError **       error)
{
    NMSettingIPConfig *s_ip6           = NULL;
    char *             value           = NULL;
    char *             method          = NM_SETTING_IP6_CONFIG_METHOD_AUTO;
    GSList *           ip6addr_list    = NULL;
    GSList *           ip6routes_list  = NULL;
    char *             default_gateway = NULL;
    add_routes_info_t  routes_info;
    ip_conf_cb_info_t  info = {NULL, AF_INET6};
    gboolean           config_ipv6;

    g_return_val_if_fail(options != NULL, NULL);
    g_return_val_if_fail(iface_path != NULL, NULL);

    config_ipv6 = options_true_value(options, "CONFIG_IPV6", FALSE);
    if (!config_ipv6) {
        method = NM_SETTING_IP6_CONFIG_METHOD_IGNORE;
        goto done;
    }

    value = options_get_value(options, "BOOTPROTO");
    if (value) {
        if (!g_ascii_strncasecmp(value, "static", 6))
            /* Assume static configuration, but it can be RA
			 * if there is no a configuration.
			 * Check it later. */
            method = NM_SETTING_IP6_CONFIG_METHOD_MANUAL;
        else if (!g_ascii_strncasecmp(value, "dhcp6", 5))
            method = NM_SETTING_IP6_CONFIG_METHOD_DHCP;
    }
    g_free(value);

    if (strcmp(method, NM_SETTING_IP4_CONFIG_METHOD_MANUAL) != 0)
        goto done;

    info.data = &ip6addr_list;
    if (for_each_line(iface_path, "ipv6address", add_ip_conf_addr_cb, &info, error)
        && ip6addr_list) {
        /* Reverse the list for matching with order in config file. */
        ip6addr_list = g_slist_reverse(ip6addr_list);
    } else
        goto error;

    routes_info.p_routes_list   = &ip6routes_list;
    routes_info.default_gateway = NULL;
    info.data                   = &routes_info;
    if (for_each_line(iface_path, "ipv6route", add_ip_conf_route_cb, &info, error)) {
        /* Reverse the list for matching with order in config file. */
        if (ip6routes_list)
            ip6routes_list = g_slist_reverse(ip6routes_list);
        default_gateway = routes_info.default_gateway;
    } else if (g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
        /* Don't care if ipv6route file does not exist */
        g_clear_error(error);
    } else
        goto error;

    /* If there is no configuration assime it is RA.
	 * There is no such method like RA-only in NM, so
	 * set it to NM_SETTING_IP6_CONFIG_METHOD_AUTO. */
    if (!ip6routes_list && !ip6addr_list)
        method = NM_SETTING_IP6_CONFIG_METHOD_AUTO;

done:
    s_ip6 = (NMSettingIPConfig *) nm_setting_ip6_config_new();
    if (!s_ip6) {
        PLUGIN_SET_ERROR(error, "Not enough memory for NMSettingIP6Config");
        goto error;
    }

    g_object_set(s_ip6, NM_SETTING_IP_CONFIG_METHOD, method, NULL);

    if (config_ipv6) {
        if (ip6addr_list) {
            g_slist_foreach(ip6addr_list, ip_settings_add_address_cb, s_ip6);
            g_slist_free_full(ip6addr_list, (GDestroyNotify) nm_ip_address_unref);
            ip6addr_list = NULL;
        }

        if (default_gateway) {
            g_object_set(s_ip6, NM_SETTING_IP_CONFIG_GATEWAY, default_gateway, NULL);
            g_free(default_gateway);
            default_gateway = NULL;
        }

        if (ip6routes_list) {
            g_slist_foreach(ip6routes_list, ip_settings_add_route_cb, s_ip6);
            g_slist_free_full(ip6routes_list, (GDestroyNotify) nm_ip_route_unref);
            ip6routes_list = NULL;
        }

        /* Read resolv.conf */
        info.data = s_ip6;
        if (!for_each_line(iface_path, "resolv.conf", set_ip_conf_dns_cb, &info, error)
            && g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
            /* Don't care if resolv.conf file does not exist */
            g_clear_error(error);
        }
    }

    return NM_SETTING(s_ip6);

error:
    if (ip6addr_list)
        g_slist_free_full(ip6addr_list, (GDestroyNotify) nm_ip_address_unref);
    if (default_gateway)
        g_free(default_gateway);
    if (ip6routes_list)
        g_slist_free_full(ip6routes_list, (GDestroyNotify) nm_ip_route_unref);
    if (s_ip6)
        g_object_unref(s_ip6);
    return NULL;
}

/***********************************************/
static gboolean
get_mtu(iplink_info_t *info, int *mtu)
{
    long int tmp = -1;

    g_return_val_if_fail(info != NULL, FALSE);
    g_return_val_if_fail(mtu != NULL, FALSE);

    if (info->mtu && get_int(info->mtu, &tmp)) {
        if (tmp >= 0 && tmp < 65536) {
            *mtu = tmp;
            return TRUE;
        } else
            PLUGIN_WARNING("Invalid MTU '%s'", info->mtu);
    }

    return FALSE;
}

static NMSetting *
make_wired_setting(const char *iface_path, GError **error)
{
    NMSettingWired *s_wired;
    int             mtu;
    iplink_info_t * info;

    s_wired = NM_SETTING_WIRED(nm_setting_wired_new());

    info = get_iplink_info(iface_path, error);
    if (info) {
        if (get_mtu(info, &mtu))
            g_object_set(s_wired, NM_SETTING_WIRED_MTU, mtu, NULL);

        if (info->mac) {
            g_object_set(s_wired, NM_SETTING_WIRED_CLONED_MAC_ADDRESS, info->mac, NULL);
        }
    }

    iplink_info_free(info);
    return (NMSetting *) s_wired;
}

static NMConnection *
read_wired_connection(const char *    iface_path,
                      options_file_t *options,
                      GError **       error)
{
    NMConnection *connection    = NULL;
    NMSetting *   con_setting   = NULL;
    NMSetting *   wired_setting = NULL;

    g_return_val_if_fail(iface_path != NULL, NULL);
    g_return_val_if_fail(options != NULL, NULL);

    connection = nm_simple_connection_new();
    if (!connection) {
        PLUGIN_SET_ERROR(error, "Failed to allocate new connection for %s.", iface_path);
        return NULL;
    }

    con_setting = make_connection_setting(iface_path,
                                          options,
                                          NM_SETTING_WIRED_SETTING_NAME,
                                          NULL);
    if (!con_setting) {
        PLUGIN_SET_ERROR(error, "Failed to create connection setting.");
        goto error;
    }
    nm_connection_add_setting(connection, con_setting);

    wired_setting = make_wired_setting(iface_path, error);
    if (!wired_setting)
        goto error;

    nm_connection_add_setting(connection, wired_setting);

    if (!nm_connection_verify(connection, error)) {
        goto error;
    }

    return connection;

error:

    g_object_unref(connection);
    if (con_setting)
        g_object_unref(con_setting);
    if (wired_setting)
        g_object_unref(wired_setting);
    return NULL;
}

static char *
get_supplicant_config_value(GHashTable *wsconf, const char *key, gboolean to_lower)
{
    char *value;
    g_return_val_if_fail(wsconf != NULL, NULL);
    g_return_val_if_fail(key != NULL, NULL);

    value = g_hash_table_lookup(wsconf, key);
    if (!value)
        return NULL;
    return to_lower ? g_ascii_strdown(value, -1) : g_strdup(value);
}

static GSList *
get_supplicant_config_value_list(GHashTable *wsconf, const char *key, gboolean to_lower)
{
    char *  values;
    gchar **str_array;
    int     i;
    GSList *list = NULL;

    values = get_supplicant_config_value(wsconf, key, to_lower);

    if (!values)
        return NULL;
    str_array = g_strsplit_set(values, " \t", 0);
    if (str_array) {
        for (i = 0; str_array[i] != NULL; ++i) {
            list = g_slist_prepend(list, g_strdup(str_array[i]));
        }

        g_strfreev(str_array);
    }

    return list;
}

enum wireless_mode { WL_MODE_INFRASTRUCTURE, WL_MODE_ADHOC, WL_MODE_AP, WL_MODE_UNKNOWN };

static enum wireless_mode
get_wireless_mode(GHashTable *wsconf)
{
    char *             mode_str;
    enum wireless_mode mode = WL_MODE_INFRASTRUCTURE;

    g_return_val_if_fail(wsconf != NULL, WL_MODE_UNKNOWN);

    mode_str = get_supplicant_config_value(wsconf, "mode", FALSE);

    if (mode_str) {
        if (!g_ascii_strcasecmp(mode_str, "0"))
            mode = WL_MODE_INFRASTRUCTURE;
        else if (!g_ascii_strcasecmp(mode_str, "1"))
            mode = WL_MODE_ADHOC;
        else if (!g_ascii_strcasecmp(mode_str, "2"))
            mode = WL_MODE_AP;
        else
            mode = WL_MODE_UNKNOWN;
    }

    g_free(mode_str);

    return mode;
}

static char *
get_ssid(GHashTable *wsconf)
{
    return get_supplicant_config_value(wsconf, "ssid", FALSE);
}

static char *
get_bssid(GHashTable *wsconf)
{
    return get_supplicant_config_value(wsconf, "bssid", FALSE);
}

static char *
get_key_mgmt(GHashTable *wsconf)
{
    char *key_mgmt;

    key_mgmt = get_supplicant_config_value(wsconf, "key_mgmt", TRUE);
    if (!key_mgmt)
        key_mgmt = g_strdup("none");

    return key_mgmt;
}

static guint32
get_wep_tx_keyidx(GHashTable *wsconf)
{
    long int wep_tx_keyidx = 0;
    char *   wep_tx_keyidx_str;

    wep_tx_keyidx_str = get_supplicant_config_value(wsconf, "wep_tx_keyidx", TRUE);
    if (wep_tx_keyidx_str) {
        if (!get_int(wep_tx_keyidx_str, &wep_tx_keyidx) || wep_tx_keyidx < 0 || wep_tx_keyidx > 3) {
            wep_tx_keyidx = 0;
        }
    }

    return (guint32) wep_tx_keyidx;
}

static char *
get_auth_alg(GHashTable *wsconf)
{
    return get_supplicant_config_value(wsconf, "auth_alg", TRUE);
}

static GSList *
get_proto(GHashTable *wsconf)
{
    return get_supplicant_config_value_list(wsconf, "proto", TRUE);
}

static GSList *
get_pairwise(GHashTable *wsconf)
{
    return get_supplicant_config_value_list(wsconf, "pairwise", TRUE);
}

static GSList *
get_group(GHashTable *wsconf)
{
    return get_supplicant_config_value_list(wsconf, "group", TRUE);
}

static char *
get_wep_key(GHashTable *wsconf, int n)
{
    char * key_str = NULL;
    char * val     = NULL;
    char * wep_key = NULL;
    size_t len;

    key_str = g_strdup_printf("wep_key%d", n);

    val = get_supplicant_config_value(wsconf, key_str, FALSE);
    if (!val)
        goto out;

    len = strlen(val);

    if (len == 10 || len == 26) {
        /* Hexadecimal WEP key */
        char *p = val;

        while (*p) {
            if (!g_ascii_isxdigit(*p)) {
                PLUGIN_WARNING("Invalid hexadecimal WEP key.");
                goto out;
            }
            p++;
        }
    } else if (len == 5 || len == 13) {
        /* ASCII passphrase */
        char *p = val;

        while (*p) {
            if (!isascii((int) (*p))) {
                PLUGIN_WARNING("Invalid ASCII WEP passphrase.");
                goto out;
            }
            p++;
        }
    } else {
        PLUGIN_WARNING("Invalid WEP key length.");
        goto out;
    }

    wep_key = g_strdup(val);

out:

    g_free(key_str);
    g_free(val);

    return wep_key;
}

static char *
get_psk(GHashTable *wsconf)
{
    char *value;

    value = get_supplicant_config_value(wsconf, "psk", FALSE);
    if (!value)
        return NULL;
    if (strlen(value) > 64 || strlen(value) < 8) {
        PLUGIN_WARNING("Error loading psk: not between 8 and 64 characters inclusive");
        g_free(value);
        return NULL;
    }

    return value;
}

static NMSetting *
make_wireless_setting(GHashTable *wsconf,
                      const char *iface_path,
                      GError **   error)
{
    NMSettingWireless *s_wireless;
    char *             str;
    enum wireless_mode mode;
    int                mtu;
    iplink_info_t *    info;

    s_wireless = NM_SETTING_WIRELESS(nm_setting_wireless_new());

    str = get_ssid(wsconf);
    if (str) {
        gsize len = strlen(str);

        if (len > 0 && len <= 32) {
            GBytes *ssid;

            ssid = g_bytes_new(str, len);
            g_object_set(s_wireless, NM_SETTING_WIRELESS_SSID, ssid, NULL);
            g_bytes_unref(ssid);
        } else {
            PLUGIN_SET_ERROR(error,
                             "Invalid SSID '%s' (size %zu not between 1 and 32 inclusive)",
                             str,
                             len);
            g_free(str);
            goto error;
        }
        g_free(str);
    }

    mode = get_wireless_mode(wsconf);
    switch (mode) {
    case WL_MODE_INFRASTRUCTURE:
        g_object_set(s_wireless, NM_SETTING_WIRELESS_MODE, "infrastructure", NULL);
        break;
    case WL_MODE_ADHOC:
        g_object_set(s_wireless, NM_SETTING_WIRELESS_MODE, "adhoc", NULL);
        break;
    case WL_MODE_AP:
        g_object_set(s_wireless, NM_SETTING_WIRELESS_MODE, "ap", NULL);
        break;
    default:
        PLUGIN_SET_ERROR_LITERAL(error, "Invalid mode");
    }

    info = get_iplink_info(iface_path, error);
    if (info) {
        if (get_mtu(info, &mtu)) {
            g_object_set(s_wireless, NM_SETTING_WIRELESS_MTU, mtu, NULL);
        }
        if (info->mac) {
            g_object_set(s_wireless, NM_SETTING_WIRELESS_CLONED_MAC_ADDRESS, info->mac, NULL);
        }
    }

    iplink_info_free(info);

    str = get_bssid(wsconf);
    if (str) {
        g_object_set(s_wireless, NM_SETTING_WIRELESS_BSSID, str, NULL);
        g_free(str);
    }

    return NM_SETTING(s_wireless);

error:
    if (s_wireless)
        g_object_unref(s_wireless);
    return NULL;
}

static char *
get_printable_ssid(NMSetting *setting)
{
    GBytes *ssid;
    char *  printable_ssid = NULL;

    ssid = nm_setting_wireless_get_ssid(NM_SETTING_WIRELESS(setting));
    if (ssid)
        printable_ssid =
            nm_utils_ssid_to_utf8(g_bytes_get_data(ssid, NULL), g_bytes_get_size(ssid));

    return printable_ssid;
}

static void
set_proto_cb(gpointer data, gpointer user_data)
{
    NMSettingWirelessSecurity *s_wireless_sec = (NMSettingWirelessSecurity *) user_data;

    g_return_if_fail(data != NULL);
    g_return_if_fail(s_wireless_sec != NULL);

    /* NetworkManager don't understand 'WPA2', replace with 'RSN' */
    if (!g_ascii_strcasecmp((const char *) data, "WPA2")) {
        g_free(data);
        data = g_strdup("rsn");
    }
    nm_setting_wireless_security_add_proto(s_wireless_sec, (const char *) data);
    g_free(data);
}

static void
set_pairwise_cb(gpointer data, gpointer user_data)
{
    NMSettingWirelessSecurity *s_wireless_sec = (NMSettingWirelessSecurity *) user_data;

    g_return_if_fail(data != NULL);
    g_return_if_fail(s_wireless_sec != NULL);

    nm_setting_wireless_security_add_pairwise(s_wireless_sec, (const char *) data);
    g_free(data);
}

static void
set_group_cb(gpointer data, gpointer user_data)
{
    NMSettingWirelessSecurity *s_wireless_sec = (NMSettingWirelessSecurity *) user_data;

    g_return_if_fail(data != NULL);
    g_return_if_fail(s_wireless_sec != NULL);

    nm_setting_wireless_security_add_group(s_wireless_sec, (const char *) data);
    g_free(data);
}

static NMSetting *
make_wireless_security_setting(GHashTable *wsconf, NMSettingWireless *s_wireless, GError **error)
{
    NMSettingWirelessSecurity *s_wireless_sec;
    char *                     auth_alg = get_auth_alg(wsconf);
    char *                     str;
    GSList *                   list;
    int                        i;

    /*FIXME: make_leap_settings() not implemented */
    // 	if (auth_alg && !g_ascii_strcasecmp (str, "leap"))
    // 		return make_leap_settings(wsconf, error);

    s_wireless_sec = NM_SETTING_WIRELESS_SECURITY(nm_setting_wireless_security_new());

    str = get_key_mgmt(wsconf);
    if (str) {
        g_object_set(s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_KEY_MGMT, str, NULL);
        g_free(str);
    }

    g_object_set(s_wireless_sec,
                 NM_SETTING_WIRELESS_SECURITY_WEP_TX_KEYIDX,
                 get_wep_tx_keyidx(wsconf),
                 NULL);

    if (auth_alg) {
        g_object_set(s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_AUTH_ALG, auth_alg, NULL);
        g_free(auth_alg);
    }

    list = get_proto(wsconf);
    if (list) {
        g_slist_foreach(list, set_proto_cb, s_wireless_sec);
        g_slist_free(list);
    }

    list = get_pairwise(wsconf);
    if (list) {
        g_slist_foreach(list, set_pairwise_cb, s_wireless_sec);
        g_slist_free(list);
    }

    list = get_group(wsconf);
    if (list) {
        g_slist_foreach(list, set_group_cb, s_wireless_sec);
        g_slist_free(list);
    }

    for (i = 0; i <= 3; ++i) {
        str = get_wep_key(wsconf, i);
        if (str) {
            nm_setting_wireless_security_set_wep_key(s_wireless_sec, i, str);
            g_free(str);
        }
    }

    str = get_psk(wsconf);
    if (str) {
        g_object_set(s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_PSK, str, NULL);
        g_free(str);
    }

    return (NMSetting *) s_wireless_sec;
}

static NMConnection *
read_wireless_connection(const char *    iface_path,
                         options_file_t *options,
                         GHashTable *    wsconf,
                         GError **       error)
{
    NMConnection *connection = NULL;
    NMSetting *   setting;
    NMSetting *   security;
    char *        printable_ssid;

    g_return_val_if_fail(iface_path != NULL, NULL);
    g_return_val_if_fail(options != NULL, NULL);
    g_return_val_if_fail(error != NULL, NULL);
    g_return_val_if_fail(*error == NULL, NULL);

    connection = nm_simple_connection_new();
    if (!connection) {
        PLUGIN_SET_ERROR(error, "Failed to allocate new connection for %s.", iface_path);
        return NULL;
    }

    if (!wsconf) {
        g_object_unref(connection);
        PLUGIN_SET_ERROR(error, "Wireless settings are not found in %s.", iface_path);
        return NULL;
    }
    setting = make_wireless_setting(wsconf, iface_path, error);
    if (!setting) {
        g_object_unref(connection);
        return NULL;
    }

    nm_connection_add_setting(connection, setting);
    security = make_wireless_security_setting(wsconf, NM_SETTING_WIRELESS(setting), error);
    if (security) {
        nm_connection_add_setting(connection, security);
    }
    printable_ssid = get_printable_ssid(setting);
    setting        = make_connection_setting(iface_path,
                                      options,
                                      NM_SETTING_WIRELESS_SETTING_NAME,
                                      printable_ssid);
    nm_connection_add_setting(connection, setting);
    g_free(printable_ssid);

    return connection;
}

/* Only NM_DEVICE_TYPE_WIFI and NM_DEVICE_TYPE_ETHERNET are
 * recognized */
static NMDeviceType
get_device_type(options_file_t *options, const char *iface_name)
{
    NMDeviceType dev_type        = NM_DEVICE_TYPE_UNKNOWN;

    g_return_val_if_fail(options != NULL, NM_DEVICE_TYPE_UNKNOWN);

    if (options->type == NULL)
        return NM_DEVICE_TYPE_UNKNOWN;

    if (!g_ascii_strcasecmp("eth", options->type)) {
        /* Try to get device link type */
        int ifindex = nm_platform_link_get_ifindex(NM_PLATFORM_GET, iface_name);
        if (ifindex > 0) {
            dev_type = nm_platform_link_get_type(NM_PLATFORM_GET, ifindex) == NM_LINK_TYPE_WIFI ?
                       NM_DEVICE_TYPE_WIFI : NM_DEVICE_TYPE_ETHERNET;
        } else {
            dev_type = options_true_value(options, "CONFIG_WIRELESS", FALSE) ?
                       NM_DEVICE_TYPE_WIFI : NM_DEVICE_TYPE_ETHERNET;
        }
    }

    return dev_type;
}

NMConnection *
read_connection(const char * iface_path,
                GHashTable * wsconf,
                gboolean *   ignored,
                GError **    error)
{
    NMConnection *       connection = NULL;
    options_file_t *     options    = NULL;
    NMSetting *          s_ip4 = NULL, *s_ip6 = NULL;
    char *               iface_name = NULL;
    NMDeviceType         dev_type = NM_DEVICE_TYPE_UNKNOWN;
    NMSettingConnection *s_con;
    gboolean             nm_controlled;

    g_return_val_if_fail(iface_path != NULL, NULL);
    g_return_val_if_fail(ignored != NULL, NULL);

    *ignored = FALSE;

    iface_name = get_iface_name(iface_path);
    if (!iface_name) {
        PLUGIN_SET_ERROR(error, "Invalid connection dir '%s'.", iface_path);
        goto done;
    }

    options = open_options_file(iface_path);
    if (!options) {
        PLUGIN_SET_ERROR(error,
                         "Couldn't parse file '%s/options'",
                         iface_path);
        goto done;
    }

    /* if 'options' file for interface is exists and
	 * NM_CONTROLLED is not set then interface should be
	 * ignored, so set default to FALSE.
	 * If 'options' file is not exists then interface
	 * should be managed, so set default to TRUE.
	 * Otherwise NM_CONTROLLED value will be read from
	 * 'options' file. */
    nm_controlled = options_true_value(options, "NM_CONTROLLED", options->options == NULL);

    dev_type = get_device_type(options, iface_name);

    /* If 'options' file for interface is present then
	 * ignore interface if its type is unsupported */
    if (options->options && dev_type == NM_DEVICE_TYPE_UNKNOWN) {
        /* If NM_CONTROLLED is explicitly set to 'yes' then
         * don't ignore interface even if its type is unknown */
        *ignored = nm_controlled ? FALSE : TRUE;
        goto done;
    }

    if (!nm_controlled) {
        *ignored = TRUE;
        goto done;
    }

    switch (dev_type) {
    case NM_DEVICE_TYPE_ETHERNET:
        connection = read_wired_connection(iface_path, options, error);
        break;
    case NM_DEVICE_TYPE_WIFI:
        connection =
            read_wireless_connection(iface_path, options, wsconf, error);
        break;
    default:
        PLUGIN_SET_ERROR(error,
                         "Connection for %s : Unknown device type",
                         iface_name);
        break;
    }

    if (!connection)
        goto done;

    s_ip4 = make_ip4_setting(options, iface_path, error);
    if (!s_ip4) {
        g_object_unref(connection);
        connection = NULL;
        goto done;
    } else {
        nm_connection_add_setting(connection, s_ip4);
    }

    s_ip6 = make_ip6_setting(options, iface_path, error);
    if (!s_ip6) {
        g_object_unref(connection);
        connection = NULL;
        goto done;
    } else {
        nm_connection_add_setting(connection, s_ip6);
    }

    /* Don't allow to use this connection for other interfaces */
    s_con = nm_connection_get_setting_connection(connection);
    if (!s_con) {
        PLUGIN_SET_ERROR(error, "Can't get connection settings for %s", iface_name);
        g_object_unref(connection);
        connection = NULL;
        goto done;
    } else {
        g_object_set(s_con, NM_SETTING_CONNECTION_INTERFACE_NAME, iface_name, NULL);
    }

    if (!nm_connection_verify(connection, error)) {
        PLUGIN_WARNING("Connection for %s invalid: %s", iface_name, (*error)->message);
        g_object_unref(connection);
        connection = NULL;
    }

done:
    g_free(iface_name);
    close_options_file(options);

    return connection;
}
